#!/usr/local/bin/perl
#
# KTH-whois++ v1.1 alpha
# A prototype whois++ server in perl
# Rickard Schoultz <schoultz@admin.kth.se>
# Aug 1993, KTH/SUNET NOC
#
#################### Start of configurable options #######################

$serverhandle="KTHSE02";	# Handle for this server 
                                # (IANA handle in some way in some future)
$BASEDIR="/home/staff/schoultz/work/whois++"; 
                                # Base directory
$DBDIR="$BASEDIR/dbdir";	# Directory to hold databases
$LOGFile="$BASEDIR/LOG";	# Log file

#################### End of configurable options #######################

#require "conf.pl";
require "getopts.pl";
require "timelocal.pl";

sub usage {			# Print Usage
    print <<"EOT";

Usage: indexserver.pl options

Server options:
   -S [-p port]    For normal server operation (default port is 43)
   -I              For server mode run from command line or inetd
		   (Not recommended).
Database options: 
   -M files        To make the database (eg. indexserver.pl -M dbdir/*)

Network options:
   -D [-h host -p port | -s serverhandle]
                   For issuing a DATA-CHANGED to another server
   -P -h host [-p port] [-t template] [-f field]
                   For polling another server
   -R              Reset the polled-by database
   -C              Clear the local centroids database

EOT
}

$0="Whois++ Server";
$version="1.1a";
$CFILE="$DBDIR/.Centroids";	# Centroid file
$DBName="$DBDIR/.Lookup";
$MAPName="$DBDIR/.Offsets";
$pollers="$DBDIR/.Pollers";
$polled="$DBDIR/.Polled";

%abridgedview = (
 "user", "name handle email",
 "organization", "organization-name handle organization-city",
 "services", "name handle protocol",
 "siteinfo", "name handle city",
 "help", "template handle subject",
 "command", "command",
 "template", "handle"
		 );
$hostname=`hostname`; chop $hostname;
open (LOG, ">>$LOGFile") || warn "Could not write log file";

do Getopts('s:h:p:SIDPMRC');
if (defined $opt_p) {
    $port=$opt_p;
} else {
    $port=43;
}
if (defined $opt_D) {
#    require "$BASEDIR/datachanged.pl";
    &issue_data_changed($opt_s,$opt_h,$opt_p);
    exit;
} elsif (defined $opt_P) {
    require "$BASEDIR/poll.pl";
    &poll($opt_h,$opt_p,$opt_t,$opt_f);
    exit;
} elsif (defined $opt_M) {
    require "$BASEDIR/makedb.pl";
    &makedb;
    exit;
} elsif (defined $opt_R) {
    &resetpolls;
    exit;
} elsif (defined $opt_C) {
    &resetcentroids;
    exit;
} elsif (!defined $opt_S && !defined $opt_I) {
    &usage;
    exit;
}

if ($opt_I) {
    $stdin=STDIN;
    $stdout=STDOUT;
    $0.=" (inetd)";
} else {
    $0.=" (Waiting)";
    &setupserver($port);
    print "% KTH-whois++ server version:$version. Type 'help' for help.\n";
}

dbmopen(DB, "$DBName", undef) || die "Database is rebuilding, please try again in a while";
open(MAP, "$MAPName") || die "Database is rebuilding, please try again in a while";
dbmopen(CEN, "$DBDIR/.Centroids", undef) || eval {$nocentroids=1;};

$|=1;

$globalmode="";
$globalmaxhits=200;

 INPUT:
    $input=<$stdin>;
$input=~s/\cM//g;
$input=~s/^[ \t]*//;
chop $input;
&log($input);
$_=$input;
$input=~s/:[ \t]*[Hh][Oo][Ll][Dd]$// if ($hold=/:[ \t]*hold$/i);

($command, @args) = split(/ /, $input);
$_=$command;
tr/A-Z/a-z/;

$maxhits=$globalmaxhits;
$mode=$globalmode;

FOO: {
    &help(@args),last FOO if (/^help/ || /^\?/);
    &list(@args),last FOO if (/^list/);
    &show(@args),last FOO if (/^show/);
    &constraints(@args),last FOO if (/^constraints/);
    &version(@args),last FOO if (/^version/);
    &describe(@args),last FOO if (/^describe/);
    &commands(@args),last FOO if (/^commands/);
    &polled_by(@args),last FOO if (/^polled-by/);
    &polled_for(@args),last FOO if (/^polled-for/);
    &holdit(@args),last FOO if (/^hold/);
    &forwarded_query(), last FOO if /^forwarded-query/;
    &got_data_changed(), last FOO if /^data-changed/;
    &polled(),last FOO if /^poll/;
    &exit(0) if ((/^quit/) || (/^$/));
    &search($input);
}
goto INPUT if ($hold || $globalhold);
&exit(0);

sub getconstraint {		# Extract constraints from command line
    local($constraints)=@_;
    local($unmatched);
    $unmatched="";
    foreach $constraint (split (/,/, $constraints)) {
	$_=$constraint;
      LOOP1: {
	  /format=/ && do {
	      s/format=(.*)/\1/;
	      if (/full|abridged|handle|summary/) {
		  $mode=$_;
	      } 
	      last LOOP1;};
	  
	  /maxhits=/ && do {
	      s/maxhits=(.*)/\1/;
	      $maxhits=$_;
	      last LOOP1;};
	  $unmatched.=$constraint;
      }
    }
    return $unmatched;
}

sub getspecifier {		# What data will be searched?
    local ($input)=@_;
    local ($tmp);
    local ($svalue, $sattribute, $method,$template);
    $_=$input;
  FOOBAR: {
      /^\*/ && do {		
          s/^.//;
          $svalue=$_;
          $method="value";
          last FOOBAR;};
      /^\^/ && do {		
          s/^.//;
          $svalue=$_;
	  $sattribute="template";
          $method="attribute-value";
          last FOOBAR;};
      /^\./ && do {		
          s/^.//;
          $svalue=$_;
	  $sattribute="attribute";
          $method="attribute-value";
          last FOOBAR;};
      /^!/ && do {
          s/^.//;
	  $svalue=$_;
	  $sattribute="handle";
	  $method="attribute-value";
	  last FOOBAR;};
      /^\#/ && do {		
          s/^.//;
          $svalue=$_;
          $method="value";
          last FOOBAR;};
      /value=/ && do {
	  ($sattribute,$svalue) = split(/=/);
	  $method="value";    
	  last FOOBAR;};
      /maxhits=/ && do {	# Yes, a constraint here, but it's global.
	  ($tmp, $globalmaxhits) = split(/=/);
	  $maxhits=$globalmaxhits;
	  $method="constraint";
	  last FOOBAR;};
      /format=/ && do {		# Yes, a constraint here, but it's global.
	  ($tmp, $globalmode) = split(/=/);
	  $mode=$globalmode;
	  $method="constraint";
	  last FOOBAR;};
      /={1,1}/ && do {
	  ($sattribute,$svalue) = split(/=/);
	  $method="attribute-value";
	  last FOOBAR;};
      $svalue=$_;
      $method="value";
  }
    $sattribute=~s/template/template-type/;
    return ($method, $sattribute, $svalue);
}

sub specialterm {		# Maybe search term is special
    local ($input) = @_;	# like x@y etc...
    $_=$input;
  LOOP: {
      /@/ && !/email=/ && do {
	  $input="email=$input";
	  last LOOP;};
  }
    return $input;
}

sub search {			# Search starting point
    local ($input,$scope)=@_;
    local ($mode,$sum);

    local ($mode,$lcount,@RESULT)=&searchlocal($input,());
    local ($mode,$ccount,@CRESULT)=&searchcentroid($input,()) 
	if ($scope ne "local" && !$nocentroids);
    $sum=$lcount+$ccount;
    @starter=&printstarter($mode,$sum,@summary);
    print @starter;
    print @RESULT;
    print @CRESULT if ($scope ne "local" && !$nocentroids);
    print "\n# END\n";
}

sub searchlocal {		# Search our local database
    local ($input,@RESULT)=@_;
    local ($tmp,$i,$terms,$term,$match,$method,@newlist,@oldoffsets,
	   @starter,$listcount);
    $input=~tr/A-Z/a-z/;
    $input=~s/[ \t]+//g;
    if ($input =~ /,/ ) {
	($terms,$constraints)=split(/,/, $input,2);
	&getconstraint($input);
    } else {
	$terms=$input;
    }
    $listcount=0;
    ($method,@newlist) = &iter_terms($terms); # Iterate over search terms
    if ($method eq "constraint") { # This wasn't a search, actually
	print "% Ok\n";
	last;
    } 
    if ($#newlist >= 0) {
	undef %summary;
	$listcount=$#newlist+1;	    
	# Set apropriate format constraint if user didn't specify...
	#  That is, if there is no global format already specified
	if ($mode !~ /summary|abridged|full|handle|pointer|mime/i) {
	    if ($listcount > 20 ) {
		$mode="summary"; 
	    } elsif ($listcount > 9 && $listcount <= 20) {
		$mode="abridged";
	    } else {
		$mode="full";
	    }
	}
	@RESULT=&compilematches($mode,@newlist);
    }
    return($mode,$listcount,@RESULT);
}

sub compilematches {		# Get records for matches
    local ($mode,@newlist)=@_;
    local ($matchcount);
    foreach $i (0..$#newlist) {
	local ($template,$attribute,$dbfile,$offset)=split(/,/,$newlist[$i]);
	$handle =~ tr/a-z/A-Z/;
	open(RAW, "$DBDIR/$dbfile") if ($dbfile ne $olddbfile);
	$olddbfile=$dbfile;
	seek(RAW, $offset,0);
	while(<RAW>) {
	    last if /^$/;
	    if (/^[^ \t]/) {
		$attribute=$_;
		$attribute=~s/(.+):.*$/\1/;
		$attribute=~tr/A-Z/a-z/;
		chop $attribute;
	    }
	    if (/^handle/i) {
		chop;
		tr/a-z/A-Z/;
		s/[ 	]//g;
		($tmp,$handle)=split(/:/);
	    } elsif ( $attribute !~ /^template-type/i) {
		&addfield($template, $attribute, $_, $mode);
	    }
	}
	if ($mode =~ /^a/i) {
	    $fieldlist{"handle"} = $handle;
	    $fieldlist{"template"} = $template;
	}
	if ($maxhits == 0 || $matchcount++ < $maxhits) {
	    @RESULT=&writerecord($template, $handle, $mode);
	}
    }
    return @RESULT;
}

sub printstarter {		# Print First lines in a search response
    local ($mode,$sum,@summary)=@_;
    local (@starter);
    $mode="full" if ($sum < 1);
    $_=$mode;
  FOO: {
      /^f/i && do {
	  push(@starter, "# FULL $sum\n\n");
	  last FOO;};
      /^a/i && do {
	  push(@starter, "# ABRIDGED $sum\n\n");
	  last FOO;};
      /^s/i && do {
	  push(@starter, "# SUMMARY\n\n");
	  push(@starter," Matches:   $sum\n");
	  $once="";
	  foreach $a (keys %summary) {
	      if (!$once) {
		  $once=1;
		  push(@starter, " Templates: ${a}\n");
	      } else {
		  push(@starter, "            ${a}\n");
	      }
	  }
	  last FOO;};
      /^h/i && do {
	  push(@starter, "# HANDLE $sum\n\n");
	  last FOO;};
  }
    return @starter;
}

sub iter_terms {		# Iterates and matches over terms
    local ($terms)=@_;
    local ($term,$method,$sattribute,$searchstring,$dbmatch,$value);
    @newlist=@oldoffsets=@offsets=();
    foreach $term (split(/;/,$terms)) {
	$term=&specialterm($term);
	($method, $sattribute,$searchstring) = &getspecifier($term);
	last if ($method eq "constraint");
	
	@offsets=@newlist=();
	last if (($dbmatch=$DB{$searchstring}) == 0);
	&skiplines(MAP,$dbmatch);
	$_=<MAP>; chop;
	($word,$value)=split(/:/,$_,2);
	&match_terms;
	# Remove those that doesn't match the previous matches (AND)
	if (@oldoffsets) {
	    # and_list should return @newlist and @offsets
	    &and_list;
	}
	@oldoffsets=@offsets;
    }
    return ($method,@newlist);
}
    
sub match_terms {		# Match search term to attributes
#    ($value,@newlist,@offsets)=@_;  # Perl4:Can't be done
    local($template,$attribute,$dbfile,$offset,$match);
    foreach $match (split(/\|/,$value)) {
	($template,$attribute,$dbfile,$offset)=split(/,/,$match);
	if ((($method eq "value") ||
	     ($attribute eq $sattribute)) && 
	    (!grep(/$offset/, @offsets))) {
	    push(@newlist, $match) ;
	    push(@offsets, $dbfile.$offset);
	}
    }
#   return(@newlist,@offsets);   perl4:can't be done!
}
	
sub and_list {			# AND two match lists
#    (@newlist,@offsets,@oldoffsets)=@_;  # Perl4:Can't be done
    local(@andlist,@newoffsets);
    foreach $i (0..$#offsets) {
	if (grep(/$offsets[$i]/, @oldoffsets)) {
	    push(@andlist,$newlist[$i]);
	    push(@newoffsets,$offsets[$i]);
	}
    }
    @newlist=@andlist;
    @offsets=@newoffsets;
#   return(@newlist,@offsets);   perl4:can't be done!
}

sub writerecord {		# Push record
    local ($template, $handle, $mode) = @_;
    local ($i,$tmp);
    $_=$mode;
  KAKA: {
      /^f/i && do {
	  $template=~tr/a-z/A-Z/;
	  push(@RESULT,"# $template $handle\n");
	  foreach $i (0..$#fieldlist) {push(@RESULT, " $fieldlist[$i]\n");};
	  last KAKA;};
      /^a/i && do {
	  foreach $a (split (/ /,$abridgedview{$template})) {
	      $tmp.=sprintf ("%-20.30s ", $fieldlist{$a});
	  }
	  push(@RESULT,$tmp);
	  push(@RESULT,"\n");
	  last KAKA;};
      /^s/i && do {
	  $summary{$template}++ if ($template !~ /template/i);
	  last KAKA;};
      /^h/i && do {
	  push (@RESULT, sprintf (" %-10s %s\n", $handle, $template));
	  last KAKA;};
  }
    @fieldlist=();
    return @RESULT;
}

sub addfield {			# Push field
    local ($template, $attribute, $field, $mode) = @_;
    $_=$mode;
  BAR: {
      /^f/i && do {
	  chop $field;
	  push (@fieldlist,$field);
	  last BAR;};
      /^a/i && do {
	  if ($abridgedview{$template} =~ /$attribute/) {
	      $field=~s/.+:[ \t]*(.*)/\1/;
	      chop $field;
	      $fieldlist{$attribute}=$field;
	  }
	  last BAR;};
  }
}

sub skiplines {			# "seek(file,line)"
    local($filehandle,$lines)=@_;
    local($tmp);
    seek($filehandle,0,0);
    $doit="<$filehandle>";
    for (1..$lines) {
	eval $doit;
    }
}

sub forwarded_query {		# Got forwarded-query command
    local(@fquery,$qkey,$value,$line);
    @fquery=&readblock($stdin);
    foreach $qkey (keys %fquery) {
	&log ("FORWARDED FROM ".$fquery{"forwarded-from"}.
	      " : $qkey:$fquery{$qkey}");
    }
    &search($fquery{"body-of-query"});
}

sub readblock {			# Read a "record oriented" command
    local($stdin)=@_;
    while (<$stdin>) {
	s/\cM//g;
	last if (/^end/i || /^$/);
	chop;
	$line=$_;
	$line=~tr/A-Z/a-z/;
	($qkey,$value)=split(/:/,$line,2);
	$qkey=~s/[ 	]//g;
	$fquery{$qkey}=$value;
    }
    return @fquery;
}

sub searchcentroid {		# Search our collection of centroids
    local ($input,@CRESULT) = @_;
    local ($lc,$count,@LIST,@TMP,@search_list,$a,$mode);
    ($searchkey,$constraints)=split(/,/,$input,2);
    if (defined $CEN{"$searchkey"}) {
	foreach $a (split(/\|/,$CEN{$searchkey})) {
	    ($server,$template,$field)=split(/:/,$a);
	    $search_list{"$server"}=1;
	}
    }
    foreach $a (keys %search_list) {
	($address,$port)=&handle2address($a);
	($mode,$lc,@TMP)=&forward_query($address,$port,$input);
	$count+=$lc;
	push (@LIST,@TMP);
    }
    return ($mode, $count,@LIST);
}

sub forward_query {		# Forward a query to another server
    local ($host,$port,$input)=@_;
    local(@TMP);
    require "sys/socket.ph";  
    ($status,$msg)=&connect_to($host,$port,"S");
    if (!$status) {
	print "% Error on $host port $port : $msg\n"; # Should handle this
	return ();
    }
    &makeforward($input);
    @TMP=&getforwardanswer;
    close(S);	
    return @TMP;
}

sub makeforward {		# make forward command to send
    local($input)=@_;
    print S "FORWARDED-QUERY:\n";
    print S "  Version-Number:$version\n";
    print S "  Forwarded-From:$serverhandle\n";
    print S "  Forwarded-Time:".&time2human(time)."\n";
    print S "  Trace-option:No\n";
    print S "  Query-originating-address:$remoteaddr\n"; # NOT true! but... 
    print S "  Body-of-query:$input\n";	
    print S "END FORWARDED-QUERY\n";
}

sub getforwardanswer {		# Parse response after forwarded-query
    local(@CRESULT,$listcount);
    local($tmp,$mode);
    while(<S>) {
	next if (/^$|^%|^\#[ \t]+END.*/i);
	if (/^\#/) {
	    if (/full|abridged|handle|summary|mime|pointer/i) {
		($tmp,$mode,$listcount)=split(/[ \t]/);
		next;
	    }
	}
	push(@CRESULT,$_);
    }
    return($mode,$listcount,@CRESULT);
}

sub help {			# Help command
    local($arg)=@_;
    $arg="help" if ($arg eq "");
    &search("template=help;subject=$arg", "local");
}

sub list {			# List command
    &search("template=template,format=full", "local");
}

sub show {			# Removed from protocol...???
    local ($n)=@_;
    &search("handle=$n,template=template", "local");
}

sub constraints {		# Constraints command
    &search("template=constraint,format=full", "local");
}

sub version {			# Version command
    print "% This is KTH-Whois++ server Version 0.1.\n";
    print "# FULL 1\n";
    print "# VERSION\n";
    print " Version: $version\n";
    print "# END\n";  
}

sub describe {			# Describe command
    &search("handle=$serverhandle", "local");
}

sub commands {			# Commands command
    &search("template=command,format=full", "local");
}

sub polled_by {
    print &get_polled_by;
}

sub get_polled_by {			# Polled-by command
    local($a,$b,$count,@RESULT);
    $count=0;
    dbmopen(POLL,"$pollers",undef);
    foreach (keys %POLL) {$count++;};
    push (@RESULT,"# FULL $count\n\n");
    while (($a,$b) = (each %POLL)) {
	local ($handle,$template,$field)=split(/:/,$a);
	local ($name,$addr,$start,$end)=split(/:/,$b);
	push (@RESULT,"# POLLED-BY $handle\n");
	push (@RESULT,"  Server-Handle:$handle\n");
	push (@RESULT,"  Template:$template\n");
	push (@RESULT,"  Field:$field\n");
    }
    push (@RESULT,"\n# END\n");
    return(@RESULT);
}

sub polled_for {		# Polled-for command
    local(@tmp);
    open(POLLED,"$polled");
    @tmp=<POLLED>;
    print "# FULL ".($#tmp+1)."\n\n";   
    foreach $a (0..$#tmp) {
	($handle,$template,$field)=split(/:/,$tmp[$a]);
	print "# POLLED-FOR $handle\n";
	print "  Server-Handle:$handle\n";
	print "  Template:$template\n";
	print "  Field:$field\n";
    }
    print "\n# END\n";
}

sub holdit {			# Hold command
    $hold=1;$globalhold=1;
    print "% Holding connection.\n";
}

sub log {			# Log argument to log file
    local($input) = @_;
    ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst)=localtime(time);
    $today=sprintf ("%02D%02D%02D %02D:%02D:%02D", $year, $mon, $mday,
		    $hour, $min, $sec);
    print LOG "$today $$ $input\n";
}

sub issue_data_changed {	# Send Data-changed to another server
    local ($handle,$host,$port) = @_;
    if ($handle) {
	($host,$port)=&handle2address($handle);
	print "WHAT:$host,$port,$handle.\n";
    }
    local ($status,$msg)=&connect_to($host,$port,"S");
    if (!$status) {
	print "% Error on $host port $port : $msg\n"; # Should handle this
	return ();
    }
    &makedatachanged($input);
    @TMP=&getanswer;
    close(S);	
    return @TMP;
}

sub makedatachanged {		# Make data-changed command
    local($input)=@_;
    print S "DATA-CHANGED\n";
    print S "  Version-Number:$version\n";
#    print S "  Time-latest-centroid-change:".&time2human(time)."\n";
    print S "  Time-of-message-generation:".&time2human(time)."\n";
    print S "  Server-handle:$serverhandle\n";
    print S "  Best-time-to-poll:\n";
    print S "END FORWARDED-QUERY\n";
}

sub getanswer {			# Necessary?
    while(<S>) {
	print;
    }
}
# This routine will probably be replaced by some other method to get
# data from directory-of-servers.
sub handle2address {		# Get host-port from server handle
    local ($handle) =@_;
    local (@h2a,$mode,$listcount,@list);
    
    ($mode,$listcount,@list)=&searchlocal("handle=$handle",());
    foreach (@list) {
	chop;
	s/^[ \t]*(.*)[ \t]*:[ \t]*(.*)$/\1:\2/;
	tr/A-Z/a-z/;
	($key,$value)=split(/:/);
	$h2a{$key}=$value;
    }				
    if ($h2a{"host-address"} =~ /^\s*((\d+\.){3}\d+)\s*$/) {
	return($h2a{"host-address"},$h2a{"host-port"});
    } else {
	return($h2a{"host-name"},$h2a{"host-port"});
    }
}

sub gm2time {			# Parse date to integer
    local($gm) = @_;		# Format is: [day ,] [D]D MON YY HH:MM[:SS]
    local(@month);
    %month = ("jan", 0,"feb", 1,"mar", 2,"apr", 3,
	      "may", 4,"jun", 5,"jul", 6,"aug", 7,
	      "sep", 8,"oct", 9,"nov",10,"dec",11);
    ($day,$gm)=split(/,/, $gm) if ($gm =~ /,/);	# Do not care about day
    ($mday,$monthname,$year,$hours,$min)=split(/[: ]/, $gm);
    $time = &timegm($sec,$min,$hours,$mday,$month{$monthname},$year);
    return ($time);
}

sub time2human {		# Make a human readable date string
    local($time) = @_;
    local(@month);
    @month = ("Jan","Feb","Mar","Apr","May","Jun",
	      "Jul","Aug","Sep","Oct","Nov","Dec");
    local ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst)=gmtime($time);
    $time=sprintf("%d %s %2d %.2d\:%.2d\:%.2d GMT", $mday, $month[$mon], 
		  $year,$hour,$min,$sec);
    return("$time");
}

sub setupserver {		# Setup server on this host
    local($port)=@_;
    local($AF_INET,$SOCK_STREAM,$name,$aliases,$proto,$sockaddr,$this,
	  $con,$child,$af,$addrtype,$length,@inetaddr);
    $AF_INET = 2;
    $SOCK_STREAM = 1;
    ($name, $aliases, $proto) = getprotobyname('tcp');
    if ($port !~ /^\d+$/) {
	($name, $aliases, $port) = getservbyport($port, 'tcp');
    }
    $sockaddr = 'S n a4 x8';
    $this = pack($sockaddr, $AF_INET, $port, "\0\0\0\0");
    select(NS); $| = 1; select(STDOUT);
    socket(S, $AF_INET, $SOCK_STREAM, $proto) || die "socket: $!";
    bind(S, $this) || die "bind: $!";
    listen(S, 5) || die "connect: $!";
    select(S); $| = 1; select(STDOUT);
    for ($con = 1; ; $con++) {
	if ($addr = accept(NS, S)) {
	    if ($child = fork) {
		waitpid($child, 0);
	    } else {
		if (fork) {
		    close(NS);
		    exit;
		} else {
		    ($af, $port, $inetaddr) = unpack($sockaddr, $addr);
		    @inetaddr = unpack('C4', $inetaddr);
		    ($remotename, $aliases, $addrtype, $length, @addrs) = 
			gethostbyaddr($inetaddr, $AF_INET);
		    $inetaddr = join('.', @inetaddr);
		    $0="- $remotename (Whois++)";
		    &log ("Connect:$remotename ($inetaddr).");
		    $stdin=NS;
		    select(NS);
		    return;
		}
	    }
	}
    }
}

sub connect_to {		# Open a connection to another server
  local ($Server, $Port, $Socket_Handle) = @_;
  $AF_INET = 2;
  $SOCK_STREAM = 1;
  $IP_EXPR = '^\s*((\d+\.){3}\d+)\s*$';
  $PORT_EXPR = '^\s*(\d+)\s*$';
  $SOCKADDR_TEMPLATE = 'S n a4 x8';
  $LISTEN_QUEUE = 5;
  if ($Server =~ /$IP_EXPR/) {
     $Server_Addr = pack ('C4', split (/\./, $1));
  } else {
     $Server_Addr = (gethostbyname ($Server))[4];
     return (0, "no such host $Server") if (! length ($Server_Addr));
  }
  $TCP_Protocol_Num = (getprotobyname "TCP")[2];
  return (0, "no such protocol TCP") if ($TCP_Protocol_Num eq "");
  ($Status, $Msg, $Port) = &Service_Or_Port_To_Port ($Port);
  return (0, $Msg) if (! $Status);
  $My_Name = pack ($SOCKADDR_TEMPLATE, $AF_INET, 0, "\0\0\0\0");
  $Server_Name = pack ($SOCKADDR_TEMPLATE, $AF_INET, $Port, $Server_Addr);
  return (0, $!) 
     if (! socket ($Socket_Handle, $AF_INET, $SOCK_STREAM, $TCP_Protocol_Num));
  return (0, $!) if (! bind ($Socket_Handle, $My_Name)); 
  return (0, $!) if (! connect ($Socket_Handle, $Server_Name));
  $Present_Handle = select ($Socket_Handle);  
  $| = 1;   
  select ($Present_Handle);
  return (1);
}

sub Service_Or_Port_To_Port {	# Return service
   local ($Port) = @_;
   local ($Service);
 
   # If the port is a service name look it up
   if ($Port =~ /$PORT_EXPR/)
   {
      return (1, "", $1);
   }
   else
   {
      $Service = $Port;
      $Port = (getservbyname ($Service, $PROTOCOL))[2];
      return (0, "no such service $Service") if (! $Port);
      return (1, "", $Port);
   };
 
};

sub got_data_changed {		# Server got data-changed command
    local(@fquery,$host,$port,$handle);
    @fquery=&readblock($stdin); 
    foreach $qkey (keys %fquery) {
 	&log("DATA-CHANGED ".$fquery{"Server-Handle"}.  	 
	     " :$qkey:$fquery{$qkey}");
    } 
    $handle=$fquery{"server-handle"};
    ($host,$port)=&handle2address($handle); 
    if ($host eq "") { 	
	print "% That server ($handle) is not ehem.. registered\n";
	return;
    }
    $port=43 if ($port==0);
    if (fork) {
	return;
    } else {
	# This one polls almost unconditionally... (shudder)
	# (but not a server that is not known to it (dbdir/Servers)).
	&poll($handle,$host,$port,"ALL","name");
	exit;
    }
}

sub polled {			# Server got polled
    local($starttime,$endtime,$qkey,$value,$line);
    local(@centroid,@record,@pq),
    dbmopen(POLL,"$pollers",0644) || warn "Can't write POLL database";
    while (<$stdin>) {
	s/\cM//g;
	last if (/^end poll/i || /^$/);
	chop;
	$line=$_;
	$line=~tr/A-Z/a-z/;
	($qkey,$value)=split(/:/,$line,2);
	$qkey=~s/[ 	]//g;
	$pq{$qkey}=$value;
	
    }
    foreach $qkey (keys %pq) {
	&log ("POLL FROM ".$pq{"server-handle"}.
	      " : $qkey:$pq{$qkey}");
    }
    $POLL{"$pq{'server-handle'}:$pq{'template'}:$pq{'field'}"}=
	"$remotename:$inetaddr:$pq{'version-number'}"
	    .":$pq{'start-time'}:$pq{'end-time'}";
    $starttime=&gm2time($pq{"start-time"}) if (defined $phquery{"start-time"});
    $endtime=&gm2time($pq{"end-time"}) if (defined $phquery{"start-time"});;
    @centroid=&makecentroid($pq{"template"}, $pq{"field"});
    $endtime=time;
    &centroidchanges($starttime,$endtime,"FULL",@centroid);
}

sub makecentroid {		# Construct a centroid
    local($ctemplate,$cfield) = @_;
    local ($server,$word,$template,$attribute,$offset,@centroid,@temp);
    seek(MAP,0,0);
    # Get centroid from local database
    while(<MAP>) {
	chop;
	($word,$_)=split(/:/,$_,2);
	foreach (split(/\|/)) {
	    ($template,$attribute,$dbfile,$offset)=split(/,/);
	    if (!defined $temp{$word}) { 
		if (($ctemplate eq "all") || ($ctemplate eq $template)) {
		    if (($cfield eq "all") || ($cfield eq $attribute)) {
			push(@centroid, "$template:$attribute:$word");
			$temp{$word}=1;
			next;
		    }
		}
	    }
	} 
    }
    # Now add to centroid from other centroids
    if (!$nocentroids) {
	while (($word,$value) = each (%CEN)) {
	    if (!defined $temp{$word}) {
		foreach $a (split(/\|/, $value)) {
		    ($server,$template,$field)=split(/:/,$a);
		    if (($ctemplate eq "all") || ($ctemplate eq $template)) {
			if (($cfield eq "all") || ($cfield eq $field)) {
			    push(@centroid, "$template:$field:$word");
			    $temp{$word}=1;
			    next;
			}
		    }
		}
	    }		
	}
    }
    return @centroid;
}

sub centroidchanges {		# Give a centroid-changes response
    local($starttime,$endtime,$operation,@centroid) = @_;
    local($template,$field,$value,$n);
    
    $n=@centroid;
    print "# FULL 1\n";
    print "# CENTROID-CHANGES\n";
    print "  Version-number:$version\n";
    print "  Start-time:$starttime\n";
    print "  End-Time:".&time2human($endtime)."\n";
    print "  Server-Handle:$serverhandle\n";
    print "  Authentication-type:NONE\n";
    print "  Authentication-data:\n";
    print "  Compression-type:NONE\n";
    print "  Size-of-compressed-data\n";
    print "  Operation:$operation\n";
    for (sort @centroid) {
	($template,$field,$value) = split(/:/);
	$template=~ tr/a-z/A-Z/;
	print "  Template:$template\n  Field:$field\n  Data:$value\n";
    }
    print "  END CENTROID-CHANGES\n";
    print "\n# END\n";
}

sub poll {			# Send poll to another server
    local($handle, $host,$port,$template,$field)=@_;

    $template="ALL" unless ($template);
    $field="ALL" unless ($field);

    dbmopen(CEN,"$CFILE",0644) || 
	die "Could not open/create centroid database file";

    local ($status,$msg)=&connect_to($host,$port,"S");
    die $msg if (!$status);
    &makepoll;
    &getcentroid;
    open(POLLED,">$polled") || warn "Can't write POLLED database";
    print POLLED "$handle:$template:$field\n";
    close(S);	
}

sub makepoll {			# Make poll command
    print S "POLL\n";
    print S "Start-Time:\n";
    print S "End-Time:\n";
    print S "Template:$template\n";
    print S "Field:$field\n";
    print S "Server-Handle:$serverhandle\n";
    print S "Authentication-type:NONE\n";
    print S "END POLL\n";
}

sub getcentroid {		# Parse gotten centroid after poll
    local($string,$sh,$template,$field);
    while (<S>) {
	chop;
	if (!/^[\#%]/) {
	    tr/A-Z/a-z/;
	    s/^[ \t]*(.*)[ \t]*:[ \t]*(.*)/\1:\2/;
	    ($key,$value) = split(/:/,$_,2);
	    if ($key =~ /server-handle/i) {
		$sh=$value;
	    } elsif ($key =~ /template/i) {
		$template=$value;
	    } elsif ($key =~ /field/i) {
		$field=$value;
	    } elsif ($key =~ /version-number|start-time|end-time|authentication|compression|size-of-compressed-data|operation|end centroid|^$/i) {
		next;		
	    } else {
		$string="$sh:$template:$field";
		if (defined $CEN{$value}) {
		    if ($string !~ /$CEN{$value}/) {
			$CEN{$value}.="|$string";
		    }
		} else {
		    $CEN{$value}="$string";
		}
	    }
	}
	return if (/\# end/i);
    }
}

sub resetpolls {		# Clear polled-by database
    unlink($pollers.".dir");
    unlink($pollers.".pag");
}
sub resetcentroids {		# Clear our collection of centroids
    unlink($CFILE.".dir");
    unlink($CFILE.".pag");
}

sub exit {			# Exit client connection
    &log("Closing $con");
    close(NS);
    close($stdin);
    exit;
}

